@@ -1,3 +1,6 @@
+from __future__ import annotations
+
+
 __copyright__ = """
 Copyright (C) 2009-2017 Andreas Kloeckner
 Copyright (C) 2014-2017 Aaron Meurer
@@ -26,10 +29,49 @@ THE SOFTWARE.
 import os
 import sys
 from configparser import ConfigParser
+from functools import partial
+from typing import TYPE_CHECKING, Literal, TypedDict, cast
 
 from pudb.lowlevel import get_breakpoint_invalid_reason, lookup_module, settings_log
 
 
+if TYPE_CHECKING:
+    from bdb import Breakpoint
+    from collections.abc import Sequence
+
+    from urwid import CheckBox
+
+    from pudb.debugger import DebuggerUI
+    from pudb.var_view import Stringifier, VarAccessLevel
+
+
+class ConfDict(TypedDict):
+    shell: str
+    theme: str
+    line_numbers: bool
+    seen_welcome: str
+    sidebar_width: float
+    variables_weight: float
+    stack_weight: float
+    breakpoints_weight: float
+    current_stack_frame: Literal["top", "bottom"]
+    stringifier: Stringifier
+    custom_theme: str
+    custom_stringifier: str
+    custom_shell: str
+    wrap_variables: bool
+    default_variables_access_level: VarAccessLevel
+    display: str
+    prompt_on_quit: bool
+    hide_cmdline_win: bool
+    cmdline_height: float
+    hotkeys_code: str
+    hotkeys_variables: str
+    hotkeys_stack: str
+    hotkeys_breakpoints: str
+    hotkeys_toggle_cmdline_focus: str
+
+
 # see https://github.com/inducer/pudb/pull/453 for context
 _home = os.environ.get("HOME", os.path.expanduser("~"))
 XDG_CONF_RESOURCE = "pudb"
@@ -64,7 +106,7 @@ SAVED_BREAKPOINTS_FILE_NAME = "saved-breakpoints-%d.%d" % sys.version_info[:2]
 BREAKPOINTS_FILE_NAME = "breakpoints-%d.%d" % sys.version_info[:2]
 
 
-_config_ = [None]
+_config_: list[ConfDict | None] = [None]
 
 
 def load_config():
@@ -78,20 +120,21 @@ def load_config():
 
     cparser = ConfigParser()
 
-    conf_dict = {}
+    conf_dict: ConfDict = {}  # pyright: ignore[reportAssignmentType]
     try:
         cparser.read([
             join(cdir, XDG_CONF_RESOURCE, CONF_FILE_NAME)
             for cdir in XDG_CONFIG_DIRS if isdir(cdir)])
 
         if cparser.has_section(CONF_SECTION):
-            conf_dict.update(dict(cparser.items(CONF_SECTION)))
+            conf_dict.update(cast("ConfDict",
+                                  cast("object", dict(cparser.items(CONF_SECTION)))))
     except Exception:
         settings_log.exception("Failed to load config")
 
     conf_dict.setdefault("shell", "internal")
     conf_dict.setdefault("theme", "classic")
-    conf_dict.setdefault("line_numbers", "False")
+    conf_dict.setdefault("line_numbers", False)
     conf_dict.setdefault("seen_welcome", "a")
 
     conf_dict.setdefault("sidebar_width", 0.5)
@@ -107,14 +150,14 @@ def load_config():
     conf_dict.setdefault("custom_stringifier", "")
     conf_dict.setdefault("custom_shell", "")
 
-    conf_dict.setdefault("wrap_variables", "True")
+    conf_dict.setdefault("wrap_variables", True)
     conf_dict.setdefault("default_variables_access_level", "public")
 
     conf_dict.setdefault("display", "auto")
 
-    conf_dict.setdefault("prompt_on_quit", "True")
+    conf_dict.setdefault("prompt_on_quit", True)
 
-    conf_dict.setdefault("hide_cmdline_win", "False")
+    conf_dict.setdefault("hide_cmdline_win", False)
 
     # hotkeys
     conf_dict.setdefault("hotkeys_code", "C")
@@ -123,9 +166,9 @@ def load_config():
     conf_dict.setdefault("hotkeys_breakpoints", "B")
     conf_dict.setdefault("hotkeys_toggle_cmdline_focus", "ctrl x")
 
-    def normalize_bool_inplace(name):
+    def normalize_bool_inplace(name: str):
         try:
-            if conf_dict[name].lower() in ["0", "false", "off"]:
+            if conf_dict[name].lower() in ["0", "false", "off"]:  # pyright: ignore[reportUnknownMemberType]
                 conf_dict[name] = False
             else:
                 conf_dict[name] = True
@@ -141,7 +184,7 @@ def load_config():
     return conf_dict
 
 
-def save_config(conf_dict):
+def save_config(conf_dict: ConfDict):
     # This may not raise, as it is called during import.
 
     from os.path import join
@@ -163,7 +206,7 @@ def save_config(conf_dict):
         settings_log.exception("Failed to save config")
 
 
-def edit_config(ui, conf_dict):
+def edit_config(ui: DebuggerUI, conf_dict: ConfDict):
     import urwid
 
     old_conf_dict = conf_dict.copy()
@@ -174,7 +217,7 @@ def edit_config(ui, conf_dict):
 
     def _update_line_numbers():
         for sl in ui.source:
-            sl._invalidate()
+            sl._invalidate()  # pyright: ignore[reportPrivateUsage]
 
     def _update_prompt_on_quit():
         pass
@@ -196,9 +239,13 @@ def edit_config(ui, conf_dict):
     def _update_wrap_variables():
         ui.update_var_view()
 
-    def _update_config(check_box, new_state, option_newvalue):
+    def _update_config(
+                check_box: CheckBox,
+                new_state: bool,
+                option_newvalue: tuple[str, str | None]
+            ):
         option, newvalue = option_newvalue
-        new_conf_dict = {option: newvalue}
+        new_conf_dict: ConfDict = cast("ConfDict", cast("object", {option: newvalue}))
         if option == "theme":
             # only activate if the new state of the radio button is 'on'
             if new_state:
@@ -207,7 +254,7 @@ def edit_config(ui, conf_dict):
                     lb.set_focus(lb_contents.index(theme_edit_list_item))
                     return
 
-                conf_dict.update(theme=newvalue)
+                conf_dict.update({"theme": newvalue})
                 _update_theme()
 
         elif option == "line_numbers":
@@ -238,13 +285,14 @@ def edit_config(ui, conf_dict):
                     lb.set_focus(lb_contents.index(stringifier_edit_list_item))
                     return
 
-                conf_dict.update(stringifier=newvalue)
+                conf_dict["stringifier"] = cast("Stringifier", newvalue)
                 _update_stringifier()
 
         elif option == "default_variables_access_level":
             # only activate if the new state of the radio button is 'on'
             if new_state:
-                conf_dict.update(default_variables_access_level=newvalue)
+                conf_dict.update(
+                    default_variables_access_level=cast("VarAccessLevel", newvalue))
                 _update_default_variables_access_level()
 
         elif option == "wrap_variables":
@@ -260,18 +308,21 @@ def edit_config(ui, conf_dict):
         "configurations settings will not be saved.\n")
 
     cb_line_numbers = urwid.CheckBox("Show Line Numbers",
-            bool(conf_dict["line_numbers"]), on_state_change=_update_config,
-                user_data=("line_numbers", None))
+            bool(conf_dict["line_numbers"]),
+            on_state_change=partial(
+                _update_config, option_newvalue=("line_numbers", None)))
 
     cb_prompt_on_quit = urwid.CheckBox("Prompt before quitting",
-            bool(conf_dict["prompt_on_quit"]), on_state_change=_update_config,
-                user_data=("prompt_on_quit", None))
+            bool(conf_dict["prompt_on_quit"]),
+            on_state_change=partial(
+                _update_config, option_newvalue=("prompt_on_quit", None)))
 
     hide_cmdline_win = urwid.CheckBox("Hide command line"
             f"({conf_dict['hotkeys_toggle_cmdline_focus']}) window "
                                       "when not in use",
-            bool(conf_dict["hide_cmdline_win"]), on_state_change=_update_config,
-                user_data=("hide_cmdline_win", None))
+            bool(conf_dict["hide_cmdline_win"]),
+            on_state_change=partial(
+                _update_config, option_newvalue=("hide_cmdline_win", None)))
 
     # {{{ shells
 
@@ -289,8 +340,9 @@ def edit_config(ui, conf_dict):
                 conf_dict["shell"] == name)
             for name in shells]+[
                 urwid.RadioButton(shell_rb_group, "Custom:",
-                not known_shell, on_state_change=_update_config,
-                user_data=("shell", None)),
+                not known_shell,
+                on_state_change=partial(
+                    _update_config, option_newvalue=("shell", None))),
                 shell_edit_list_item,
                 urwid.Text("\nTo use a custom shell, see examples/shell.py "
                     "in the pudb distribution. Enter the full path to a "
@@ -314,12 +366,14 @@ def edit_config(ui, conf_dict):
     theme_edit_list_item = urwid.AttrMap(theme_edit, "input", "focused input")
     theme_rbs = [
             urwid.RadioButton(theme_rb_group, name,
-                conf_dict["theme"] == name, on_state_change=_update_config,
-                user_data=("theme", name))
+                conf_dict["theme"] == name,
+                on_state_change=partial(
+                    _update_config, option_newvalue=("theme", name)))
             for name in THEMES]+[
                 urwid.RadioButton(theme_rb_group, "Custom:",
-                    not known_theme, on_state_change=_update_config,
-                    user_data=("theme", None)),
+                    not known_theme,
+                    on_state_change=partial(
+                        _update_config, option_newvalue=("theme", None))),
                 theme_edit_list_item,
                 urwid.Text("\nTo use a custom theme, see examples/theme.py in the "
                     "pudb distribution. Enter the full path to a file like it in "
@@ -338,8 +392,8 @@ def edit_config(ui, conf_dict):
     stack_rbs = [
             urwid.RadioButton(stack_rb_group, name,
                 conf_dict["current_stack_frame"] == name,
-                on_state_change=_update_config,
-                user_data=("current_stack_frame", name))
+                on_state_change=partial(
+                    _update_config, option_newvalue=("current_stack_frame", name)))
             for name in stack_opts
             ]
 
@@ -363,13 +417,14 @@ def edit_config(ui, conf_dict):
     stringifier_rbs = [
             urwid.RadioButton(stringifier_rb_group, name,
                 conf_dict["stringifier"] == name,
-                on_state_change=_update_config,
-                user_data=("stringifier", name))
+                on_state_change=partial(
+                    _update_config, option_newvalue=("stringifier", name)))
             for name in stringifier_opts
             ]+[
                 urwid.RadioButton(stringifier_rb_group, "Custom:",
-                    not known_stringifier, on_state_change=_update_config,
-                    user_data=("stringifier", None)),
+                    not known_stringifier,
+                    on_state_change=partial(
+                        _update_config, option_newvalue=("stringifier", None))),
                 stringifier_edit_list_item,
                 urwid.Text("\nTo use a custom stringifier, see "
                     "examples/stringifier.py in the pudb distribution. Enter the "
@@ -397,8 +452,9 @@ def edit_config(ui, conf_dict):
     default_variables_access_level_rbs = [
             urwid.RadioButton(default_variables_access_level_rb_group, name,
                 conf_dict["default_variables_access_level"] == name,
-                on_state_change=_update_config,
-                user_data=("default_variables_access_level", name))
+                on_state_change=partial(
+                    _update_config,
+                    option_newvalue=("default_variables_access_level", name)))
             for name in default_variables_access_level_opts
             ]
 
@@ -407,8 +463,9 @@ def edit_config(ui, conf_dict):
     # {{{ wrap variables
 
     cb_wrap_variables = urwid.CheckBox("Wrap variables",
-            bool(conf_dict["wrap_variables"]), on_state_change=_update_config,
-                user_data=("wrap_variables", None))
+            bool(conf_dict["wrap_variables"]),
+            on_state_change=partial(
+                _update_config, option_newvalue=("wrap_variables", None)))
 
     wrap_variables_info = urwid.Text("\nNote that you can change this option on "
                                      "a per-variable basis by selecting the "
@@ -439,39 +496,47 @@ def edit_config(ui, conf_dict):
     # }}}
 
     lb_contents = (
-            [heading]
-            + [urwid.AttrMap(urwid.Text("General:\n"), "group head")]
-            + [cb_line_numbers]
-            + [cb_prompt_on_quit]
-            + [hide_cmdline_win]
-
-            + [urwid.AttrMap(urwid.Text("\nShell:\n"), "group head")]
-            + [shell_info]
-            + shell_rbs
-
-            + [urwid.AttrMap(urwid.Text("\nTheme:\n"), "group head")]
-            + theme_rbs
-
-            + [urwid.AttrMap(urwid.Text("\nStack Order:\n"), "group head")]
-            + [stack_info]
-            + stack_rbs
-
-            + [urwid.AttrMap(urwid.Text("\nVariable Stringifier:\n"), "group head")]
-            + [stringifier_info]
-            + stringifier_rbs
-
-            + [urwid.AttrMap(urwid.Text("\nVariables Attribute Visibility:\n"),
-                "group head")]
-            + [default_variables_access_level_info]
-            + default_variables_access_level_rbs
-
-            + [urwid.AttrMap(urwid.Text("\nWrap Variables:\n"), "group head")]
-            + [cb_wrap_variables]
-            + [wrap_variables_info]
-
-            + [urwid.AttrMap(urwid.Text("\nDisplay driver:\n"), "group head")]
-            + [display_info]
-            + display_rbs
+            [heading,
+                urwid.AttrMap(
+                              urwid.Text("General:\n"),
+                              "group head"),
+                cb_line_numbers,
+                cb_prompt_on_quit,
+                hide_cmdline_win,
+                urwid.AttrMap(
+                              urwid.Text("\nShell:\n"),
+                              "group head"),
+                shell_info,
+                *shell_rbs,
+                urwid.AttrMap(
+                              urwid.Text("\nTheme:\n"),
+                              "group head"),
+                *theme_rbs,
+                urwid.AttrMap(
+                                          urwid.Text("\nStack Order:\n"),
+                                          "group head"),
+                stack_info,
+                *stack_rbs,
+                urwid.AttrMap(
+                              urwid.Text("\nVariable Stringifier:\n"),
+                              "group head"),
+                stringifier_info,
+                *stringifier_rbs,
+                urwid.AttrMap(
+                              urwid.Text("\nVariables Attribute Visibility:\n"),
+                              "group head"),
+                default_variables_access_level_info,
+                *default_variables_access_level_rbs,
+                urwid.AttrMap(
+                              urwid.Text("\nWrap Variables:\n"),
+                              "group head"),
+                cb_wrap_variables,
+                wrap_variables_info,
+                urwid.AttrMap(
+                              urwid.Text("\nDisplay driver:\n"),
+                              "group head"),
+                display_info,
+                *display_rbs]
             )
 
     lb = urwid.ListBox(urwid.SimpleListWalker(lb_contents))
@@ -493,7 +558,9 @@ def edit_config(ui, conf_dict):
         # Ditto for custom stringifiers
         if stringifier_rb_group[-1].state:
             newvalue = stringifier_edit.get_edit_text()
-            conf_dict.update(stringifier=newvalue, custom_stringifier=newvalue)
+            conf_dict.update(
+                    stringifier=cast("Stringifier", newvalue),
+                    custom_stringifier=newvalue)
             _update_stringifier()
 
         if shell_rb_group[-1].state:
@@ -598,7 +665,7 @@ def load_breakpoints():
     return parse_breakpoints(lines)
 
 
-def save_breakpoints(bp_list):
+def save_breakpoints(bp_list: Sequence[Breakpoint]):
     """
     :arg bp_list: a list of `bdb.Breakpoint` objects
     """
@@ -607,11 +674,11 @@ def save_breakpoints(bp_list):
         return
 
     histfile = open(get_breakpoints_file_name(), "w")
-    bp_list = {(bp.file, bp.line, bp.cond) for bp in bp_list}
-    for bp in bp_list:
+    bps = {(bp.file, bp.line, bp.cond) for bp in bp_list}
+    for bp in bps:
         line = "b %s:%d" % (bp[0], bp[1])
         if bp[2]:
-            line += ", %s" % bp[2]
+            line += f", {bp[2]}"
         line += "\n"
         histfile.write(line)
     histfile.close()
